# ------------------------------------------------------------------------------------
# Utilities for parsing the wiki page of resonances at https://lhocds.ligo-wa.caltech.edu/wiki/Resonances
# Mark Barton 1/21/14
#
# ------------------------------------------------------------------------------------
# Key functions with usage hints:
#
# dparse takes the text of the wiki page and returns a list of valid data lines parsed into columns
#  plus a list of incorrectly formatted lines (for debugging):
# 	(parseddata, baddata) = parsereswiki.dparse(rawtext)
#  	->
#  	[
#   	['H1_HIFOX', 'SUS', 'BSC1', 'ITMx_wire_rehang', 'Rigid_body_modes', 'modeV2', '2.22184 [[#quadrehang]]', '', '2.23 [[#lho3594]]', '', '']
#    	...
# 	]
# di is a dictionary mapping user-friendly names to column indices:
# 	for l in parseddata:
# 		print l[parsereswiki.di['fmeas']]
#
# rparse and ri are similar functions for the references section.
#
# ftheory, qtheory, fmeas and qmeas are comparison functions for use with sorted:
# 	fmeasorder = sorted(parseddata,key=parsereswiki.fmeas)
#
# The columns 'ftheory', 'qtheory', 'fmeas' and 'qmeas' are each a number followed by a space-separated list of refs,
#  so use split()[0] to get the number
# 	for l in parseddata:
# 		print float(l[parsereswiki.di['fmeas']].split()[0])
#
# ------------------------------------------------------------------------------------
# prerequisites:
import re # regular expression utilities

# utilities...

# function taking a pattern for one string and returning a pattern for a space-delimited list of zero or more such strings
def list(pat):
    return pat+r"?(?:\s+"+pat+r")*"

# some very basic patterns
bars = r"\|\|\s*" # typical: "|| "
word = r"(?:[\w:.-]+)" # typical: "q1:abc-123.2"
num = r"(?:[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)" # typical: "+1.23E-4"
nat = r"\d+" # typical: 3
harm = nat+r"(?:\-"+nat+r")?" # typical: "1-4"
link = r"(?:\[\[\s*#[\w\d-]+\s*\]\])" # typical: "[[ #myanchor123 ]]"
numorlink = r"(?:"+num+r"|"+link+r")"
harmorlink = r"(?:"+harm+r"|"+link+r")"

# patterns for the columns of the data table
dpatstart = r"^"+bars+r"\*\s*" # typical: "|| * "
dpatifo = bars+r"("+list(word)+r")\s*" # typical: "|| H1 H2 "
dpatgroup = bars+r"("+list(word)+r")\s*" # typical: "|| SUS SEI "
dpatloc = bars+r"("+list(word)+r")\s*" # typical: "|| BSC8 HAM3 "
dpatsys = bars+r"("+list(word)+r")\s*" # typical: "|| ITMx ETMy "
dpatsubsys = bars+r"("+list(word)+r")\s*" # typical: "|| whatever whatever_else "
dpatdesc = bars+r"([^|]*)"  # typical: "|| pretty much anything that doesn't have vertical bars "
dpatfreqtheory = bars+r"("+list(numorlink)+r")\s*" # typical: 1.2E3 [[#myref]]
dpatqtheory = bars+r"("+list(numorlink)+r")\s*" # typical: 1.2E3 [[#myref]]
dpatfreqmeas = bars+r"("+list(numorlink)+r")\s*" # typical: 1.2E3 [[#myref]]
dpatqmeas = bars+r"("+list(numorlink)+r")\s*" # typical: 1.2E3 [[#myref]]
dpatharm = bars+r"("+list(harmorlink)+r")\s*" # typical: 1-4 [[#myref]]
dpatend = bars  # typical: "|| "

# pattern for an entire data row
dpat = dpatstart+dpatifo+dpatgroup+dpatloc\
+dpatsys+dpatsubsys+dpatdesc+dpatfreqtheory\
+dpatqtheory+dpatfreqmeas+dpatqmeas+dpatharm

# compiled versions of data patterns
cdpatnum = re.compile(num)
cdpatstart = re.compile(dpatstart)
cdpat = re.compile(dpat)

# dictionary with labels for data column indices
di = {'ifo':0,'group':1,'loc':2,'sys':3,'sub':4,'desc':5,'ftheory':6,'qtheory':7,'fmeas':8,'qmeas':9,'harm':10}

# comparison functions for use in sorting
def ftheory(parsedline):
    m = cdpatnum.search(parsedline[di['ftheory']])
    return 0 if m==None else float(m.group(0))

def qtheory(parsedline):
    m = cdpatnum.search(parsedline[di['qtheory']])
    return 0 if m==None else float(m.group(0))

def fmeas(parsedline):
    m = cdpatnum.search(parsedline[di['fmeas']])
    return 0 if m==None else float(m.group(0))

def qmeas(parsedline):
    m = cdpatnum.search(parsedline[di['qmeas']])
    return 0 if m==None else float(m.group(0))

# function to parse the data section
def dparse(ll):
    
    # all lines that are flagged as data with "*" in the first column
    data = [l for l in ll if cdpatstart.match(l)]
        
    # all lines that are flagged as data but don't match pattern for valid data
    baddata = [l for l in data if not(cdpat.match(l))]

    # split the good lines into a list of strings corresponding to columns
    parseddata = [cdpat.split(l)[1:-1] for l in data if cdpat.match(l)]
    
    return (parseddata, baddata)

# patterns for the columns of the reference table    
rpatstart = r"^"+bars+r"\%\s*" # typical: "|| % "
rpatlink = bars+r"(\s*<<\s*Anchor\(\s*\w+\s*\)\s*>>\s*\[\[[^]]*\]\])\s*" # typical: "|| << Anchor( myref ) >> [[link]] "
rpatinfo = bars+r"([^|]*)"  # typical: "|| pretty much anything that doesn't have vertical bars "
rpatend = bars  # typical: "|| "

# pattern for an entire reference row
rpat = rpatstart+rpatlink+rpatinfo+rpatend

# compiled versions of reference patterns
crpatstart = re.compile(rpatstart)
crpat = re.compile(rpat)

# function to parse the reference section
def rparse(ll):
    
    # all lines that are flagged as refs with "%" in the first column
    refs = [l for l in ll if crpatstart.match(l)]
        
    # all lines that are flagged as data but don't match pattern for valid data
    badrefs = [l for l in refs if not(crpat.match(l))]

    # split the good lines into a list of strings corresponding to columns
    parsedrefs = [crpat.split(l)[1:-1] for l in refs if crpat.match(l)]
    
    return (parsedrefs, badrefs)

# dictionary with labels for reference column indices
ri = {'link':0,'info':1}